#!/usr/bin/env python
u"""
read_SLR_C20.py
Written by Tyler Sutterley (04/2022)

Reads in C20 spherical harmonic coefficients derived from SLR measurements

Dataset distributed by NASA PO.DAAC
    https://podaac-tools.jpl.nasa.gov/drive/files/GeodeticsGravity/grace/docs
        TN-05_C20_SLR.txt
        TN-07_C20_SLR.txt
        TN-11_C20_SLR.txt
        TN-14_C30_C30_GSFC_SLR.txt
Dataset distributed by UTCSR
    ftp://ftp.csr.utexas.edu/pub/slr/degree_2/C20_RL05.txt
Datasets distributed by GFZ
    ftp://isdcftp.gfz-potsdam.de/grace/Level-2/GFZ/RL06_SLR_C20/
        GFZ_RL06_C20_SLR.dat
    ftp://isdcftp.gfz-potsdam.de/grace/GravIS/GFZ/Level-2B/aux_data/
        GRAVIS-2B_GFZOP_GRACE+SLR_LOW_DEGREES_0002.dat

CALLING SEQUENCE:
    SLR_C20 = read_SLR_C20(SLR_file)

INPUTS:
    SLR_file:
        RL04: TN-05_C20_SLR.txt
        RL05: TN-07_C20_SLR.txt
        RL06: TN-11_C20_SLR.txt
        CSR: C20_RL05.txt
        GFZ: GFZ_RL06_C20_SLR.dat
        GFZ (combined): GRAVIS-2B_GFZOP_GRACE+SLR_LOW_DEGREES_0002.dat

OUTPUTS:
    data: SLR degree 2 order 0 cosine stokes coefficients (C20)
    error: SLR degree 2 order 0 cosine stokes coefficient error (eC20)
    month: GRACE/GRACE-FO month of measurement (April 2002 = 004)
    date: date of SLR measurement

OPTIONS:
    AOD: remove background De-aliasing product from the SLR solution (for CSR)
    HEADER: file contains header text to be skipped (default: True)

PYTHON DEPENDENCIES:
    numpy: Scientific Computing Tools For Python
        https://numpy.org
        https://numpy.org/doc/stable/user/numpy-for-matlab-users.html
    dateutil: powerful extensions to datetime
        https://dateutil.readthedocs.io/en/stable/

PROGRAM DEPENDENCIES:
    time.py: utilities for calculating time operations

REFERENCES:
    Cheng, M. and Tapley, B. D., "Variations in the Earth's oblateness during
        the past 28 years", Journal of Geophysical Research: Solid Earth,
        109(B9), B09402, 2004. 10.1029/2004JB003028
    Loomis, B. D., Rachlin, K. E., and Luthcke, S. B., "Improved Earth
        Oblateness Rate Reveals Increased Ice Sheet Losses and Mass-Driven Sea
        Level Rise", Geophysical Research Letters, 46(12), 6910-6917, 2019.
        https://doi.org/10.1029/2019GL082929
    Koenig, R., Schreiner, P, and Dahle, C. "Monthly estimates of C(2,0)
        generated by GFZ from SLR satellites based on GFZ GRACE/GRACE-FO
        RL06 background models." V. 1.0. GFZ Data Services, (2019).
        https://doi.org/10.5880/GFZ.GRAVIS_06_C20_SLR

UPDATE HISTORY:
    Updated 04/2022: updated docstrings to numpy documentation format
        include utf-8 encoding in reads to be windows compliant
    Updated 09/2021: use functions for converting to and from GRACE months
    Updated 05/2021: added GFZ SLR and GravIS oblateness solutions
        define int/float precision to prevent deprecation warning
    Updated 02/2021: use adjust_months function to fix special months cases
        replaced numpy bool to prevent deprecation warning
    Updated 12/2020: using utilities from time module
    Updated 08/2020: flake8 compatible regular expression strings
    Updated 07/2020: added function docstrings
    Updated 08/2019: add catch to verify input SLR file exists
    Updated 07/2019: added tilde-expansion of input SLR file
    Updated 06/2019: added new GRACE-FO special month (October 2018)
    Updated 11/2018: new TN-11 files only list GRACE months available
    Updated 06/2016: added option HEADER for files that do not have header text
    Updated 05/2016: added option AOD to not remove the AOD correction
    Updated 03/2016: minor update to read PO.DAAC
    Updated 05/2015: minor change to file determination (only regular expressions)
    Updated 02/2015: updated UT/CSR portion and comments
    Updated 09/2014: rewrite of the TN-07 read program
        using regular expressions and convert_calendar_decimal
    Updated 01/2014: updated to use UT/CSR monthly time-series
        as an alternative to PO.DAAC as it is updated more regularly
    Updated 05/2013: adapted for python
    Updated 09/2012: Changed month scheme to output.
        Used to remove the GRACE missing months in this program by feeding in the GRACE months
        BUT, as the new SLR files start with an earlier date, decided to parallel
        the degree-1 read program, and remove the missing months in the read_grace program
    Updated 06/2012: OVERHAUL of dating and modification for 'special' GRACE months
        Initiated from an incorrect date tag in the SLR data file
        New dating will convert from the MJD file into date fraction
        Some GRACE 'months' have the accelerometer turned off
            for half the month to preserve battery power
        These months use half of the prior month in the GRACE global gravity solution
        For these months the SLR file has a second dataline for the modified period
        Will use these marked (*) data to replace the GRACE C2,0
        ALSO converted the mon and slrdate inputs into options
    Updated 01/2012: Updated to feed in SLR file from outside
        Will accommodate upcoming GRACE RL05, which will use different SLR files
    Written 12/2011
"""
import warnings
import gravity_toolkit.SLR

# PURPOSE: read oblateness data from Satellite Laser Ranging (SLR)
def read_SLR_C20(*args, **kwargs):
    """
    Reads C20 spherical harmonic coefficients from SLR measurements

    Parameters
    ----------
    SLR_file: str
        Satellite Laser Ranging file
    AOD: bool, default True
        Remove background De-aliasing product from the
        Center for Space Research (CSR) SLR solutions
    HEADER: bool, default True
        File contains header text to be skipped

    Returns
    -------
    data: float
        SLR degree 2 order 0 cosine stokes coefficients
    error: float
        SLR degree 2 order 0 cosine stokes coefficient error
    month: int
        GRACE/GRACE-FO month of measurement
    date: float
        date of SLR measurement
    """
    warnings.filterwarnings("always")
    warnings.warn("Deprecated. Please use gravity_toolkit.SLR instead",
        DeprecationWarning)
    # call renamed version to not break workflows
    return gravity_toolkit.SLR.C20(*args,**kwargs)
